//==================================================================
/*

	DemoEventGenerator.cc

	Version 0.3.0	 5-Jun-1998	Sanshiro Enomoto

*/
//==================================================================

#include        <math.h>
#include        "KonoeRandomNumber.hh"
#include	"DemoEventGenerator.hh"

template <class T> T sqr(T x) {return x * x;}
template <class T> T is_same_sign(T x, T y) {return x * y > 0;}


class DemoEventVertex {
  public:
    DemoEventVertex(Konoe3Vector pincident_momentum);
    void generateParticle(Konoe3Vector &pion_momentum, Konoe3Vector &proton_momentum);
  private:
    KFloat angle2momentum(KFloat phi, KFloat pincident_momentum);
    KFloat energyViolation(KFloat p0, KFloat p, KFloat phi); // to calculate momentum
  private:
    Konoe3Vector incident_momentum;
};


class ThinMagneticField {
  public:
    ThinMagneticField(KFloat pintencity);
    void curve(Konoe3Vector &momentum, KInt charge);
  protected:
    KFloat intencity;
    KFloat width;
};


//------------------------------------------------------------------

// DemoEventGenerator definitions

//------------------------------------------------------------------

DemoEventGenerator::DemoEventGenerator()
{
    KonoeRandomNumber::initialize();
    KFloat incident_momentum = sqrt(sqr(INCIDENT_ENERGY) - sqr(PION_MASS));
    vertex = new DemoEventVertex(Konoe3Vector(0, 0, incident_momentum));
    magnetic_field = new ThinMagneticField(MAGNETIC_FIELD_INTENSITY);
}


DemoEventGenerator::~DemoEventGenerator()
{
    delete magnetic_field;
    delete vertex;
}


KonoeSimEvent *DemoEventGenerator::generate() 
{
    Konoe3Vector pion_momentum, proton_momentum;
    KonoeTrackSegment  *tracksegment[4];

    vertex->generateParticle(pion_momentum, proton_momentum);
    generateTrackSegment(tracksegment, pion_momentum, proton_momentum);

    KonoeTrack *track[2];
    track[0] = new KonoeTrack( 2 );
    track[1] = new KonoeTrack( 2 );
    track[0] -> add( tracksegment[0] );
    track[0] -> add( tracksegment[1] );
    track[1] -> add( tracksegment[2] );
    track[1] -> add( tracksegment[3] );

    KonoeSimEvent *event = new KonoeSimEvent( 2 );
    event -> add( track[0] );
    event -> add( track[1] );

    return event;
}


void DemoEventGenerator::generateTrackSegment(
    KonoeTrackSegment *tracksegment[4], 
    Konoe3Vector pion_momentum, Konoe3Vector proton_momentum
) 
{
    Konoe3Vector startpoint, endpoint;
    KFloat starttime, endtime;
    KFloat velocity;
    Konoe3Vector momentum;
    KFloat mass, energy;
    KInt charge;

  // pion track
    momentum = pion_momentum;
    mass = PION_MASS;
    energy = sqrt(sqr((KFloat) momentum) + sqr(mass));
    velocity =  LIGHT_SPEED * sqrt((1 - mass / energy) * (1 + mass / energy));
    charge = -1;
    startpoint = Konoe3Vector(0, 0, 0);
    endpoint.z = MAGNETIC_FIELD_POSITION;
    endpoint = momentum * ((endpoint.z - startpoint.z) / momentum.z);
    starttime = 0;
    endtime = (KFloat) (endpoint - startpoint) / velocity + starttime;

    tracksegment[0] = new KonoeTrackSegment( startpoint, endpoint );
    tracksegment[0] -> setStartTime( starttime );
    tracksegment[0] -> setEndTime( endtime );
    tracksegment[0] -> setVelocity( velocity );
    tracksegment[0] -> setEnergy( energy );

    magnetic_field->curve(momentum, charge);

    startpoint = endpoint;
    endpoint.z = (momentum.z > 0) ? DETECTER_END_POSITION : 0;
    endpoint = startpoint + momentum * ((endpoint.z - startpoint.z) / momentum.z);
    starttime = endtime;
    endtime = (KFloat) (endpoint - startpoint) / velocity + starttime;

    tracksegment[1] = new KonoeTrackSegment( startpoint, endpoint );
    tracksegment[1] -> setStartTime( starttime );
    tracksegment[1] -> setEndTime( endtime );
    tracksegment[1] -> setVelocity( velocity );
    tracksegment[1] -> setEnergy( energy );

  // proton track
    momentum = proton_momentum;
    mass = PROTON_MASS;
    energy = sqrt(sqr((KFloat) momentum) + sqr(mass));
    velocity = LIGHT_SPEED * sqrt((1 - mass / energy) * (1 + mass / energy));
    charge = +1;
    startpoint = Konoe3Vector(0, 0, 0);
    endpoint.z = MAGNETIC_FIELD_POSITION;
    endpoint = momentum * ((endpoint.z - startpoint.z) / momentum.z);
    starttime = 0;
    endtime = (KFloat) (endpoint - startpoint) / velocity + starttime;

    tracksegment[2] = new KonoeTrackSegment( startpoint, endpoint );
    tracksegment[2] -> setStartTime( starttime );
    tracksegment[2] -> setEndTime( endtime );
    tracksegment[2] -> setVelocity( velocity );
    tracksegment[2] -> setEnergy( energy );

    magnetic_field->curve(momentum, charge);

    startpoint = endpoint;
    endpoint.z = (momentum.z > 0) ? DETECTER_END_POSITION : 0;
    endpoint = startpoint + momentum * ((endpoint.z - startpoint.z) / momentum.z);
    starttime = endtime;
    endtime = (KFloat) (endpoint - startpoint) / velocity + starttime;

    tracksegment[3] = new KonoeTrackSegment( startpoint, endpoint );
    tracksegment[3] -> setStartTime( starttime );
    tracksegment[3] -> setEndTime( endtime );
    tracksegment[3] -> setVelocity( velocity );
    tracksegment[3] -> setEnergy( energy );
}


//------------------------------------------------------------------

// DemoEventVertex definitions

//------------------------------------------------------------------

DemoEventVertex::DemoEventVertex(Konoe3Vector pincident_momentum)
{
    incident_momentum = pincident_momentum;
}


void DemoEventVertex::generateParticle(Konoe3Vector &pion_momentum, Konoe3Vector &proton_momentum)
{
    KFloat theta, phi, p, px, py, pz;  // for scattered pion

    theta = KonoeRandomNumber::getFloat() * (2.0 * M_PI);
    phi = KonoeRandomNumber::getFloat() * ACCEPTANCE;
    p = angle2momentum(phi, incident_momentum);

    px = p * cos(theta) * sin(phi);
    py = p * sin(theta) * sin(phi);
    pz = p * cos(phi);
    
    pion_momentum = Konoe3Vector(px, py, pz);
    proton_momentum = incident_momentum - pion_momentum;
}


KFloat DemoEventVertex::energyViolation(KFloat p0, KFloat p, KFloat phi)
{
    KFloat initial_energy, pion_energy, proton_energy;

    initial_energy = sqrt(sqr(p0) + sqr(PION_MASS)) + PROTON_MASS;
    pion_energy = sqrt(sqr(p) + sqr(PION_MASS));
    proton_energy = sqrt(sqr(p * sin(phi)) + sqr(p0 - p * cos(phi)) + sqr(PROTON_MASS));

    return pion_energy + proton_energy - initial_energy;
}


KFloat DemoEventVertex::angle2momentum(KFloat phi, KFloat pincident_momentum)
{
    // calculate scatterd pion momentum by bisection method
    // assuming that just one solution exists,

    KFloat p_left = 0, p_right = pincident_momentum;
    KFloat p;
    KFloat eps = 1.0e-6;

    KFloat value_left = energyViolation(pincident_momentum, p_left, phi);
    KFloat value_right = energyViolation(pincident_momentum, p_right, phi);
    KFloat value;

    if (is_same_sign(value_left, value_right)) {
        // error
	abort();
    }
    else if (value_left == 0) {
        p = p_left;
    }
    else if (value_right == 0) {
        p = p_right;
    }
    else {
        while ((p_right - p_left) > eps) {
            p = (p_right + p_left) / 2;
	    value = energyViolation(pincident_momentum, p, phi);
	    if (fabs(value) < eps) {
	        break;
	    }

	    if (is_same_sign(value, value_left)) {
	        p_left = p;
	    }
	    else {
                p_right = p;
	    }
        }
    }

    return p;
}


//------------------------------------------------------------------

// ThinMagneticField definitions

//------------------------------------------------------------------

ThinMagneticField::ThinMagneticField(KFloat pintencity)
{
    intencity = pintencity;
    width = 1.0;
}

void ThinMagneticField::curve(Konoe3Vector &momentum, KInt charge)
{
    if (momentum.z <= 0) {
	return;
    }

    KFloat momentum_xz = sqrt(sqr(momentum.x) + sqr(momentum.z));
    KFloat radus = momentum_xz / (0.3 * intencity);
    KFloat center_x, center_z;

    if (charge > 0) {
	// rotate +90 deg on xz coordinate (see <<1>>) to get center
	center_x = +momentum.z * (radus / momentum_xz);
	center_z = -momentum.x * (radus / momentum_xz);
    }
    else {
	// rotate -90 deg on xz coordinate (see <<1>>) to get center
	center_x = -momentum.z * (radus / momentum_xz);
	center_z = +momentum.x * (radus / momentum_xz);
    }

    if (center_z + radus < width) {
	momentum.z = -momentum.z;
    }
    else {
	if (charge > 0) {
	    // calcurate exit point (see <<2>>)
	    KFloat exit_x = center_x + sqrt(sqr(radus) - sqr(width - center_z));
	    KFloat exit_z = width;

	    // rotate +90 deg on xz coordinate (see <<3>>) and rescale
	    momentum.x = -(exit_z - center_z) * (momentum_xz / radus);
	    momentum.z = +(exit_x - center_x) * (momentum_xz / radus);
	}
	else {
	    // calcurate exit point (see <<2>>)
	    KFloat exit_x = center_x - sqrt(sqr(radus) - sqr(width - center_z));
	    KFloat exit_z = width;

	    // rotate -90 deg on xz coordinate (see <<3>>) and rescale
	    momentum.x = +(exit_z - center_z) * (momentum_xz / radus);
	    momentum.z = -(exit_x - center_x) * (momentum_xz / radus);
	}
    }

    return;
}

// units:
//     Magnetic Field [T]
//     Length [m]
//     Energy, Momentum [GeV]
//     Charge [none]
//
//           |  :                                              :                   
//           |  :<----- magnetic field ----------------------->:                   
//           |  :         _                                    :                   
//           |  :        (X) B                                 :                   
//           |  :         ~                                    :                   
//           |  :                            @                 :                   
//           |  :                    @               @         :                   
//           |  :               @                         @    : <<3>>             
//           |  :           @                                 @:   rotate -90 deg  
//           |  :        @                               exit o:  @                
//           |  :     @                                      o.:      @            
//           |  :   @                                       o..:          @        
//           |  : @                                        o...:                 
//           |  @                                         o....:                   
//   (X = 0)-+--:  <<1>>                                 o.....:             
//           |@ :  o   rotate -90 deg                   o......:                   
//          @|  :     o                                o.......:                   
//        @  |  :        o                            o........:                   
//      @    |  :           o                        o...  . ..:                   
// tarck     |  :          Pxz                      o. <<2>> ..:                   
// (Q = -1)  |  :     R = ------  o                o. for this :                   
//           |  :         0.3 B      o            o.. triangle :                   
//           |  :                       o        o.............:                   
//           |  :                          o    o..............:                   
//           |  :                            o o...............:                   
//           |  :                             * <------------->:                   
//           |  :                        center                :                   
//           |  :                                              :                   
//           |  :                                              :                   
//       ----+--+-----------------------------+----------------+------------->
//           | (Z = 0)                      (Z = center_z)   (Z = width)     Z
//           |  :                                              :                   
//           |  :                                              :                   
//         X V  :                                                                  

