//===================================================================
/*

	KonoeBoxGeometry.cc

	Version 0.2.0	30-Apr-1998	H. Sakamoto
	Version 0.4.0	 1-Aug-1998	S. Inaba
	Version 0.4.1	 5-Aug-1999	T. Yamashita
		modified method "set_point"
	Version 0.5.0	21-Sep-1999	Y. Kouno
	Version 0.5.1	18-Jan-2000	H. Sakamoto
	Version 0.7.0	21-Aug-2001	T. Sumida

*/
//===================================================================

#include	<analysis/KonoeBoxGeometry.hh>
#include	<analysis/KonoeGeometryCM.hh>
#include	<KonoeCategories.hh>
#include        <iostream.h>
#include        <iomanip.h>

const KChar	* KonoeBoxGeometry::className = "KonoeBoxGeometry";

KonoeBoxGeometry::KonoeBoxGeometry( ) 
	: KonoeGeometry( KonoeGeometryCM::CID_BOXGEOMETRY ) {
}

KonoeBoxGeometry::KonoeBoxGeometry( KFloat w, KFloat h, KFloat d ) 
  : KonoeGeometry( KonoeGeometryCM::CID_BOXGEOMETRY ) {
    
	width = w;
	height = h;
	depth = d;
	waxis.set( w, 0.0, 0.0 );
	haxis.set( 0.0, h, 0.0 );
	daxis.set( 0.0, 0.0, d );
	displacement.set( 0.0, 0.0, 0.0 );
}


void	KonoeBoxGeometry::translate( KonoeTranslation * pt ) {
  
  displacement = pt -> getDisplacement( );
  waxis = pt -> rotate(  waxis );
  haxis = pt -> rotate(  haxis );
  daxis = pt -> rotate(  daxis );
}

KInt	KonoeBoxGeometry::isCrossing( KonoeTrajectory * trajectory,
				      Konoe3Vector & entrypoint, Konoe3Vector & exitpoint ) {
  
  Konoe3Vector	p, a, b;
  KFloat		fa, fb, fx;
  
  ffx = -1.0;
  
  p =  displacement;	// XY
  a =  waxis;
  b =  haxis;
  if( trajectory -> isCrossing( p, a, b, fa, fb, fx ) ){
    if( check_fx( fx ) ) {
      set_points( trajectory, entrypoint, exitpoint );
      return 1;
    }
  }
  b =  daxis;		// XZ
  if( trajectory -> isCrossing( p, a, b, fa, fb, fx ) )
    if( check_fx( fx ) ) {
      set_points( trajectory, entrypoint, exitpoint );
      return 1;
    }
  a =  haxis;
  if( trajectory -> isCrossing( p, a, b, fa, fb, fx ) )
    if( check_fx( fx ) ) {
      set_points( trajectory, entrypoint, exitpoint );
      return 1;
    }
  p =  displacement +  waxis;
  a =  haxis;
  b =  daxis;
  if( trajectory -> isCrossing( p, a, b, fa, fb, fx ) )
    if( check_fx( fx ) ) {
      set_points( trajectory, entrypoint, exitpoint );
      return 1;
    }
  p =  displacement +  daxis;
  a =  waxis;
  b =  haxis;
  if( trajectory -> isCrossing( p, a, b, fa, fb, fx ) ){
    if( check_fx( fx ) ) {
      set_points( trajectory, entrypoint, exitpoint );
      return 1;
    }
  }
  p =  displacement + haxis;
  a =  waxis;
  b =  daxis;
  if( trajectory -> isCrossing( p, a, b, fa, fb, fx ) )
    if( check_fx( fx ) ) {
      set_points( trajectory, entrypoint, exitpoint );
      return 1;
    }
  return 0;
}

KInt	KonoeBoxGeometry::check_fx( KFloat fx ) {
  
  if( ffx == -1 ) {
    ffx = fx;
    return 0;
  }
  else {
    if( fx < ffx ) {
      entrance_f = fx;
      exit_f = ffx;
    }
    else {
      entrance_f = ffx;
      exit_f = fx;
    }
    return 1;
  }
}

void	KonoeBoxGeometry::set_points( KonoeTrajectory * t,
				      Konoe3Vector & entrypoint, Konoe3Vector & exitpoint ) {
  
  Konoe3Vector	v;
  v = t -> getEndPoint( ) - t -> getStartPoint( );
  entrypoint = t->getStartPoint() + v * entrance_f;
  exitpoint = t->getStartPoint() + v * exit_f;
  
}

void	KonoeBoxGeometry::dump( ostream & os ) {
  os << setw(15) << "waxis: "; waxis.dump( os ); os << endl; 
  os << setw(15) << "haxis: "; haxis.dump( os ); os << endl; 
  os << setw(15) << "daxis: "; daxis.dump( os ); os << endl; 
  os << setw(15) << "displacement: "; displacement.dump( os ); 
  os << endl; 
  
  
}

KInt	KonoeBoxGeometry::put( KonoeOutputObjectStream *os, KChar level,
				KInt memberid ) {
	
	KChar 	l = level + 1;

	os -> putHeader( this, level, memberid, N_MEMBERS );

	os -> put( l, MID_WIDTH, width );
	os -> put( l, MID_HEIGHT, height );
	os -> put( l, MID_DEPTH,  depth );
	waxis.put( os, l, MID_WAXIS );
	haxis.put( os, l, MID_HAXIS );
	daxis.put( os, l, MID_DAXIS );
	displacement.put( os, l, MID_DISPLACEMENT );

	return 0;
	}

KInt	KonoeBoxGeometry::getStreamSize( ) {

	KInt s;

	s = HEADER_SIZE +1;
	s += HEADER_SIZE + 1;
	s += HEADER_SIZE + 1;
	s += HEADER_SIZE + waxis.getStreamSize( );
	s += HEADER_SIZE + haxis.getStreamSize( );
	s += HEADER_SIZE + daxis.getStreamSize( );
	s += HEADER_SIZE + displacement.getStreamSize( );
	return s;

	}

KInt	KonoeBoxGeometry::fillPrimitiveMember( KInt memberid, KInt datatype,
				KInt datacount, void * data ) {
	switch( datatype ) {
	case KFLOAT_TYPE :
		switch( memberid ){
		case MID_WIDTH :
			width = * ( KFloat * ) data;
			delete [] data;
			break;
		case MID_HEIGHT :
			height = * ( KFloat * ) data;
			delete [] data;
			break;
		case MID_DEPTH :
			depth = * ( KFloat * ) data;
			delete [] data;
			break;
		default :
			delete [] data;
			return -1;
			}
		return 0;
	default :
		delete [] data;
		return -1;
	}
}

KInt 	KonoeBoxGeometry::fillObjectMember( KInt memberid, 
				KonoeStreamableObject * obj ) {
	switch( memberid ) {
	case MID_WAXIS :
		waxis = * ( Konoe3Vector * ) obj;
		delete obj;
		break;
	case MID_HAXIS :
		haxis = * ( Konoe3Vector * ) obj;
		delete obj;
		break;
	case MID_DAXIS :
		daxis = * ( Konoe3Vector * ) obj;
		delete obj;
		break;
	case MID_DISPLACEMENT :
		displacement = * ( Konoe3Vector * ) obj;
		delete obj;
		break;
	default :
		return -1;
	}
	return 0;
}




